# (c) Copyright 2020 Trent Hauck
# All Rights Reserved
"""Module holding third party code."""

import json
import os.path
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Sequence
from typing import Tuple

from gcgc.tokenizer.kmer_tokenzier import KmerTokenizer

try:
    import torch
    import torch.utils.data
    from transformers.tokenization_utils import PreTrainedTokenizer

    from Bio import File
    from Bio import SeqIO

except ImportError as exp:
    # pylint: disable=invalid-name
    needed = "torch, transformers, biopython"
    raise ImportError(f"Missing one or more libraries: {needed}. Please install: {exp}") from exp


class GCGCTransformersTokenizer(PreTrainedTokenizer):
    """A GCGC Tokenizer that is compatible with the transformers library."""

    VOCABULARY_FILENAME = "gcgc_vocab.json"

    def __init__(self, alphabet, kmer_length, kmer_stride, **kwargs: Dict[str, Any]):
        """Init the GCGCTransformersTokenizer object."""
        self.kmer_tokenizer = KmerTokenizer(
            alphabet=alphabet,
            kmer_length=kmer_length,
            kmer_stride=kmer_stride,
            bos_token=kwargs.get("bos_token"),
            eos_token=kwargs.get("eos_token"),
            unk_token=kwargs.get("unk_token"),
            pad_token=kwargs.get("pad_token"),
            mask_token=kwargs.get("mask_token"),
        )

        super().__init__(
            alphabet=alphabet, kmer_length=kmer_length, kmer_stride=kmer_stride, **kwargs,
        )

        self.init_inputs = (alphabet, kmer_length, kmer_stride)
        self.init_kwargs = kwargs

    @classmethod
    def from_kmer_tokenizer(cls, kmer_tokenizer: KmerTokenizer):
        """Init from a kmer tokenizer."""
        return cls(
            kmer_tokenizer.alphabet,
            kmer_tokenizer.kmer_length,
            kmer_tokenizer.kmer_stride,
            bos_token=kmer_tokenizer.bos_token,
            eos_token=kmer_tokenizer.eos_token,
            unk_token=kmer_tokenizer.unk_token,
            pad_token=kmer_tokenizer.pad_token,
            mask_token=kmer_tokenizer.mask_token,
        )

    def get_vocab(self):
        """Return the vocabulary for this transformer."""
        return self.kmer_tokenizer.vocab.stoi

    @property
    def vocab_size(self):
        """Return the size of the vocabulary."""
        return len(self.kmer_tokenizer.vocab)

    def _convert_id_to_token(self, index: int) -> str:
        return self.kmer_tokenizer.vocab.itos[index]

    def _convert_token_to_id(self, token: str) -> int:
        return self.kmer_tokenizer.vocab.stoi[token]

    def _tokenize(self, text: str, **kwargs):
        return self.kmer_tokenizer.tokenize(text)

    def convert_tokens_to_string(self, tokens: List[str]) -> str:
        """Convert the tokens into a string."""
        return "".join(tokens)

    def save_vocabulary(self, save_directory) -> Tuple[str]:
        """Save the vocabulary string to integer map in the save_directory."""
        vocab_location = os.path.join(save_directory, self.VOCABULARY_FILENAME)

        with open(vocab_location, "w") as outf:
            json.dump(self.kmer_tokenizer.vocab.stoi, outf)

        return (vocab_location,)


# pylint: disable=too-few-public-methods
class _SequenceIndexer:
    """A helper object that is used to index multiple files at ones."""

    def __init__(self):
        """Initialize the _SequenceIndexer object."""
        self._record_index = {}
        self._counter = -1

    def __call__(self, sid):
        """Return the record index if it's known, otherwise generate a new one."""
        try:
            return self._record_index[sid]
        except KeyError:
            self._counter = self._counter + 1
            self._record_index[sid] = self._counter
            return self._record_index[sid]


class GenomicDataset(torch.utils.data.Dataset):
    """GenomicDataset can be used to load sequence information into a format aminable to PyTorch."""

    def __init__(
        self, file_index: Dict[Path, File._IndexedSeqFileDict], tokenizer: GCGCTransformersTokenizer
    ):
        """Initialize the GenomicDataset object."""
        self._file_index = file_index
        self._tokenizer = tokenizer

        super().__init__()

    @classmethod
    def from_path(
        cls, path: Path, tokenizer: GCGCTransformersTokenizer, file_format: str = "fasta",
    ) -> "GenomicDataset":
        """Init from a single file. This is a convenience method that delegates to from_paths."""
        return cls.from_paths([path], tokenizer, file_format)

    @classmethod
    def from_paths(
        cls,
        path_sequence: Sequence[Path],
        tokenizer: GCGCTransformersTokenizer,
        file_format="fasta",
    ) -> "GenomicDataset":
        """Initialize the GenomicDataset from a pathlib.Path sequence."""
        file_index = {}
        seq_indexer = _SequenceIndexer()

        for file_path in sorted(path_sequence):
            file_index[file_path] = SeqIO.index(
                str(file_path), file_format, key_function=seq_indexer
            )

        return cls(file_index, tokenizer)

    def __len__(self) -> int:
        """Return the length of the dataset."""
        return sum(len(v) for v in self._file_index.values())

    def __getitem__(self, i: int):
        """Get the record from the index."""
        for value in self._file_index.values():
            try:
                seq_record = value[i]
            except KeyError:
                continue

            tokenized = self._tokenizer.encode(str(seq_record.seq))
            # pylint: disable=not-callable, no-member
            return torch.tensor(tokenized, dtype=torch.long)

        # pylint: disable=useless-else-on-loop
        else:
            raise RuntimeError(f"Exausted file index while looking for {i}.")



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Fixtures for testing."""

import pathlib

_PATH = pathlib.Path(__file__).parent

P53_HUMAN = _PATH / "p53_human/p53_human.fasta"
ECOLI_PATH = _PATH / "ecoli"
PF01152_PATH = _PATH / "globin_alignment/PF01152_seed.sto"
PF01152_PATH_FULL = _PATH / "globin_alignment/PF01152_full.fasta"
PF12057_PATH = _PATH / "PF12057/PF12057.fasta"



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Test the alphabet."""

import pytest

from gcgc import KmerTokenizer


@pytest.mark.parametrize(
    "seq,tokenizer,expected_tokens,expected_encoding",
    [
        (
            "ATCG",
            KmerTokenizer.bare_tokenizer(alphabet="ATCG", kmer_length=1, kmer_stride=1),
            list("ATCG"),
            [0, 1, 2, 3],
        ),
        (
            "ATCG",
            KmerTokenizer.bare_tokenizer(
                alphabet="ATCG", max_length=2, kmer_length=1, kmer_stride=1
            ),
            list("AT"),
            [0, 1],
        ),
        (
            "ATCG",
            KmerTokenizer.bare_tokenizer(
                alphabet="ATCG", min_length=4, pad_token="|", kmer_length=2, kmer_stride=1
            ),
            ["AT", "TC", "CG", "|"],
            [2, 7, 12, 0],
        ),
        (
            "AATT",
            KmerTokenizer.bare_tokenizer(
                alphabet="ATCG", pad_token="|", conform_length=5, kmer_length=2, kmer_stride=2
            ),
            ["AA", "TT", "|", "|", "|"],
            [1, 6, 0, 0, 0],
        ),
        (
            "AATT",
            KmerTokenizer.bare_tokenizer(
                alphabet="ATCG",
                pad_token="|",
                conform_length=5,
                kmer_length=2,
                kmer_stride=2,
                pad_at_end=False,
            ),
            ["|", "|", "|", "AA", "TT"],
            [0, 0, 0, 1, 6],
        ),
        (
            "ATCGATCGATCGATCG",
            KmerTokenizer.bare_tokenizer(
                alphabet="ATCG", max_length=2, kmer_length=2, kmer_stride=1
            ),
            ["AT", "TC"],
            [1, 6],
        ),
        (
            "ATCGATCG",
            KmerTokenizer(
                alphabet="ATCG",
                pad_token="|",
                bos_token=">",
                eos_token="<",
                unk_token=None,
                mask_token=None,
                max_length=10,
                kmer_length=2,
                kmer_stride=1,
            ),
            [">", "AT", "TC", "CG", "GA", "AT", "TC", "CG", "<"],
            [1, 4, 9, 14, 15, 4, 9, 14, 2],
        ),
    ],
)
def test_kmer_tokenization(seq, tokenizer, expected_tokens, expected_encoding):
    """Test the kmers are encoded as expected."""
    actual_tokens = tokenizer.tokenize(seq)
    assert actual_tokens == expected_tokens

    actual_encoded = tokenizer.encode(seq)
    assert actual_encoded == expected_encoding


@pytest.mark.parametrize(
    "token_ids, tokenizer, expected_mask",
    [
        (
            [0, 1, 2, 3, 4],
            KmerTokenizer(
                alphabet="ATCG",
                kmer_length=1,
                kmer_stride=1,
                pad_token="|",
                bos_token=None,
                eos_token=None,
                unk_token=None,
                mask_token=None,
            ),
            [1, 0, 0, 0, 0],
        ),
    ],
)
def test_get_special_tokens_mask(token_ids, tokenizer, expected_mask):
    """Test the special tokens mask returns the proper mask for a given input list of token ids."""
    actual_mask = tokenizer.get_special_tokens_mask(token_ids)
    assert actual_mask == expected_mask



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Top-level GCGC module."""

import os
import warnings as _warnings

from gcgc import tokenizer
from gcgc.tokenizer.kmer_tokenzier import KmerTokenizer
from gcgc.tokenizer.kmer_tokenzier import SequenceTokenizer

__all__ = ["tokenizer", "KmerTokenizer", "SequenceTokenizer"]

__version__ = os.environ["GCGC_VERSION"]

_warnings.simplefilter(action="ignore", category=PendingDeprecationWarning)



# (c) Copyright 2020 Trent Hauck
# All Rights Reserved
"""The Vocab object holds the vocabulary for the tokenizer."""

import typing

import pydantic

# pylint: disable=unsupported-membership-test
# pylint: disable=unsupported-assignment-operation
# pylint: disable=unsubscriptable-object
# pylint: disable=no-member


class Vocab(pydantic.BaseModel):
    """The Vocab object that holds the mapping between integers and characters."""

    stoi: typing.Dict[str, int] = {}

    @property
    def itos(self):
        """Return the integer to string mapping of the underlying stoi."""
        return {integer: string for string, integer in self.stoi.items()}

    def __getitem__(self, item: str) -> int:
        """Get the item from the vocab."""
        return self.stoi[item]

    def __contains__(self, item: str) -> bool:
        """Return True if the item is in the vocab."""
        return item in self.stoi

    def __len__(self) -> int:
        """Return the size of the vocabulary."""
        return len(self.stoi)

    def __setitem__(self, item: str, value: int) -> None:
        """Add an item to the vocabulary with a specific value."""
        self.stoi[item] = value

    def add_item(self, item: str) -> None:
        """Add an item to the vocabulary."""
        try:
            current_max = max(self.stoi.values())
            self.stoi[item] = current_max + 1
        except ValueError:
            self.stoi[item] = 0

    def add_items(self, items: typing.Iterable[str]) -> None:
        """Add multiple items to the vocabulary in one fell swoop."""
        for item in items:
            self.add_item(item)



# (c) Copyright 2020 Trent Hauck
# All Rights Reserved
"""Tests for third party modules."""

import pathlib
import tempfile
from pathlib import Path

from transformers import CONFIG_MAPPING
from transformers import AutoModelWithLMHead
from transformers import DataCollatorForLanguageModeling
from transformers import Trainer
from transformers import TrainingArguments

from gcgc import KmerTokenizer
from gcgc import third_party
from gcgc.tests.fixtures import PF01152_PATH_FULL


def test_transformers_model(tmpdir):
    """Test a transformers model can be fit with a GCGC."""
    tmpdir = Path(tmpdir)

    kmer_tokenizer = KmerTokenizer(alphabet="extended_protein")
    transformers_tokenizer = third_party.GCGCTransformersTokenizer.from_kmer_tokenizer(
        kmer_tokenizer
    )

    dataset = third_party.GenomicDataset.from_path(PF01152_PATH_FULL, transformers_tokenizer)
    data_collator = DataCollatorForLanguageModeling(tokenizer=transformers_tokenizer, mlm=True)

    config = CONFIG_MAPPING["bert"](
        vocab_size=len(kmer_tokenizer.vocab),
        num_hidden_layers=2,
        hidden_size=12,
        intermediate_size=12,
        num_attention_heads=2,
        type_vocab_size=1,
    )
    model = AutoModelWithLMHead.from_config(config)

    training_args = TrainingArguments(str(tmpdir), max_steps=10)

    trainer = Trainer(
        model=model,
        args=training_args,
        data_collator=data_collator,
        train_dataset=dataset,
        prediction_loss_only=True,
    )

    trainer.train()


def test_tokenizer():
    """Test the tokenizer interface."""
    kmer_tokenizer = KmerTokenizer(alphabet="extended_protein", conform_length=9)

    transformers_tokenizer = third_party.GCGCTransformersTokenizer.from_kmer_tokenizer(
        kmer_tokenizer
    )

    with tempfile.TemporaryDirectory() as tempdir:
        temppath = pathlib.Path(tempdir)
        model_path = temppath / "model"
        model_path.mkdir(exist_ok=True)

        transformers_tokenizer.save_pretrained(str(model_path))
        transformers_tokenizer = third_party.GCGCTransformersTokenizer.from_pretrained(
            str(model_path)
        )

    # Includes start and stop ids.
    expected = [1, 15, 22, 15, 2]
    assert expected == transformers_tokenizer.encode("MVM")

    # Trim off the start and stop id.
    assert expected[1:-1] == transformers_tokenizer.convert_tokens_to_ids(list("MVM"))

    assert {
        "input_ids": [1, 15, 22, 15, 2, 0, 0, 0, 0],
        "token_type_ids": [0, 0, 0, 0, 0, 0, 0, 0, 0],
        "attention_mask": [1, 1, 1, 1, 1, 0, 0, 0, 0],
    } == transformers_tokenizer.encode_plus(
        "MVM", max_length=9, pad_to_max_length=True, truncation=True
    )

    # Test decode
    assert transformers_tokenizer.decode([1, 15, 22, 3, 2, 0]) == ">MV#<|"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gcgc',
 'gcgc.data',
 'gcgc.tests',
 'gcgc.tests.fixtures',
 'gcgc.tests.tokenizer',
 'gcgc.tests.vocab',
 'gcgc.tokenizer']

package_data = \
{'': ['*'],
 'gcgc.data': ['splice/*'],
 'gcgc.tests.fixtures': ['PF12057/*',
                         'ecoli/*',
                         'globin_alignment/*',
                         'p53_human/*']}

install_requires = \
['pydantic>=1,<2']

extras_require = \
{'third_party': ['biopython>=1,<2', 'transformers>=2', 'torch>=1.5']}

setup_kwargs = {
    'name': 'gcgc',
    'version': '0.12.3',
    'description': 'GCGC is a preprocessing library for biological sequence model development.',
    'long_description': '# GCGC\n\n> GCGC is a tool for feature processing on Biological Sequences.\n\n[![](https://github.com/tshauck/gcgc/workflows/Run%20Tests%20and%20Lint/badge.svg)](https://github.com/tshauck/gcgc/actions?query=workflow%3A%22Run+Tests+and+Lint%22)\n[![](https://img.shields.io/pypi/v/gcgc.svg)](https://pypi.python.org/pypi/gcgc)\n[![code style black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n## Installation\n\nGCGC is primarily intended to be used as part of a larger workflow inside\nPython.\n\nTo install via pip:\n\n```sh\n$ pip install gcgc\n```\n\nIf you\'d like to use code that helps gcgc\'s tokenizers integrate with common\nthird party libraries, either install those packages separately, or use gcgc\'s\nextras.\n\n```sh\n$ pip install \'gcgc[third_party]\'\n```\n\n## Documentation\n\nThe GCGC documentation is at [gcgc.trenthauck.com](http://gcgc.trenthauck.com),\nplease see it for examples.\n\n### Quick Start\n\nThe easiest way to get started is to import the kmer tokenizer, configure it,\nthen start tokenizing.\n\n```python\nfrom gcgc import KmerTokenizer\n\nkmer_tokenizer = KmerTokenizer(alphabet="unambiguous_dna")\nencoded = kmer_tokenizer.encode("ATCG")\nprint(encoded)\n```\n\nsample output:\n\n```\n[1, 6, 7, 8, 5, 2]\n```\n\nThis output includes the "bos" token, the "eos" token, and the three amino acid\ntokens in between.\n\nYou can go the other way and convert the integers to strings.\n\n```python\nfrom gcgc import KmerTokenizer\n\nkmer_tokenizer = KmerTokenizer(alphabet="unambiguous_dna")\ndecoded = kmer_tokenizer.decode(kmer_tokenizer.encode("ATCG"))\nprint(decoded)\n```\n\nsample output:\n\n```\n[\'>\', \'A\', \'T\', \'C\', \'G\', \'<\']\n```\n\nThere\'s also the vocab for the kmer tokenizer.\n\n```python\nfrom gcgc import KmerTokenizer\n\nkmer_tokenizer = KmerTokenizer(alphabet="unambiguous_dna")\nprint(kmer_tokenizer.vocab.stoi)\n```\n\nsample output:\n\n```\n{\'|\': 0, \'>\': 1, \'<\': 2, \'#\': 3, \'?\': 4, \'G\': 5, \'A\': 6, \'T\': 7, \'C\': 8}\n```\n\n### Transformers\n\nThe [Transformers](https://huggingface.co/transformers/) library has an\nidea of a tokenizer that is used for various modeling tasks.\n\nTo make it easier to use the Transformers library on biological sequences, gcgc\nhas a Transformers compatible tokenizer that can be created from the\nKmerTokenizer.\n\n```python\nfrom gcgc import KmerTokenizer\nfrom gcgc import third_party\n\nkmer_tokenizer = KmerTokenizer(\n  kmer_length=2, kmer_stride=2, alphabet="unambiguous_dna"\n)\n\ntt = third_party.GCGCTransformersTokenizer.from_kmer_tokenizer(\n    kmer_tokenizer\n)\n\nbatch = tt.batch_encode_plus(["ATGC", "GCGC"])\nprint(batch["input_ids"])\n```\n\nsample output:\n\n```\n[[1, 11, 8, 2], [1, 8, 8, 2]]\n```\n',
    'author': 'Trent Hauck',
    'author_email': 'trent@trenthauck.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'http://gcgc.trenthauck.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# (c) Copyright 2020 Trent Hauck
# All Rights Reserved
"""Test the vocabulary object."""

from typing import NamedTuple
from typing import Set

from gcgc import KmerTokenizer
from gcgc.vocab import Vocab


def test_vocab():
    """Test the vocabulary object works as expected."""
    vocab = Vocab()

    vocab.add_item("A")

    assert "A" in vocab
    assert vocab["A"] == 0

    vocab.add_items(["T", "C"])

    assert "T" in vocab
    assert "C" in vocab

    assert vocab.stoi == {"A": 0, "T": 1, "C": 2}


def test_vocab_values():
    """Test that for the alphabet the correct vocab is generated."""

    class Case(NamedTuple):
        kmer_tokenzier: KmerTokenizer
        expected: Set

    test_cases = [
        Case(KmerTokenizer.bare_tokenizer(alphabet="ATGC"), {"A", "T", "C", "G"}),
        Case(KmerTokenizer.bare_tokenizer(alphabet="ATGCN"), {"A", "T", "C", "G", "N"}),
        Case(KmerTokenizer.bare_tokenizer(alphabet="extended_dna"), set("GATCBDSW")),
        Case(KmerTokenizer.bare_tokenizer(alphabet="ambiguous_dna"), set("GATCRYWSMKHBVDN")),
    ]

    for case in test_cases:
        assert set(case.kmer_tokenzier.vocab.stoi.keys()) == case.expected, case.kmer_tokenzier
